/*
 *  Carbon_utils.c
 *  video_echo3
 *
 *  Created by David Cairns on 3/13/06.
 *  Copyright 2006 David Cairns. All rights reserved.
 *
 */

#include "Carbon_utils.h"


char GetCharForUnichar( UniChar *theUniChar )
{
	// Create a CFString from the UniChar we've passed in.
	CFStringRef inputString = CFStringCreateWithCharacters(
		NULL, 
		theUniChar, 
		1 );
	// Get a pointer to the C string for this CFString.
	const char *bytes = CFStringGetCStringPtr( 
		inputString, 
		kCFStringEncodingMacRoman );
	// Return the first character of the C string.
	return *bytes;
}


//--------------------------------------------------------------------------------------------
OSStatus SetupAppEventHandler( EventHandlerUPP appHandler, IBNibRef sNibRef, long refcon )
{
	OSStatus		err = noErr;
	static const	EventTypeSpec    kAppEvents[] = { 
		{ kCoreEventClass,		kAEQuitApplication },
		{kEventClassCommand, kEventCommandProcess} };
	
	// If the handler is NULL, just return.
	if( NULL == appHandler ) {
		return err;
	}
	
	// Create a Nib reference, passing the name of the nib file (without the .nib extension).
	// CreateNibReference only searches into the application bundle.
	err = CreateNibReference( CFSTR( "main" ), &sNibRef );
	if( noErr != err ) {
		printf( "Error creating nib reference: %i.\n", (int)err );
		return err;
	}
    
	// Once the nib reference is created, set the menu bar. "MainMenu" is the name of the menu bar
	// object. This name is set in InterfaceBuilder when the nib is created.
	err = SetMenuBarFromNib( sNibRef, CFSTR( "MenuBar" ) );
	if( noErr != err ) {
		printf( "Error creating menu bar from nib: %i.\n", (int)err );
		return err;
	}
	
	// Install our handler for "common commands" on the application target.
	err = InstallApplicationEventHandler( 
		appHandler,
		GetEventTypeCount( kAppEvents ), 
		kAppEvents,
		(void*)refcon, 
		NULL );
	if( noErr != err ) {
		printf( "Error trying to install the application event handler: %i.\n", (int)err );
	}
	
	return err;
}


OSStatus SetupKeyboardEventHandler( WindowRef winRef, EventHandlerUPP keyHandler, long refcon )
{
	OSStatus err = noErr;
	static const EventTypeSpec	kKeyEvents[] = {
		{ kEventClassTextInput, kEventTextInputUnicodeForKeyEvent } };
	
	// If the handler is NULL, just return.
	if( NULL == keyHandler ) {
		return err;
	}
	
	// Install the handler on the window target.
	err = InstallWindowEventHandler( 
		winRef, 
		keyHandler, 
		1, kKeyEvents, 
		(void*)refcon, 
		NULL );
	if( noErr != err ) {
		printf( "Error installing the keyboard event handler: %i.\n", (int)err );
		return err;
	}
	
	return err;
}


OSStatus SetupMouseEventHandler( WindowRef winRef, EventHandlerUPP mouseHandler, long refcon )
{
	// Define the events that the handler will cover.
	OSStatus err = noErr;
	
	
	static const EventTypeSpec kMouseEvents[] =  {
		{ kEventClassMouse, kEventMouseMoved }, 
		{ kEventClassMouse, kEventMouseDragged } };
	
	// If the handler is NULL, just return.
	if( NULL == mouseHandler ) {
		return err;
	}
	
	// Install the mouse handler.
	err = InstallWindowEventHandler( 
		winRef, 
		mouseHandler, 
		2, kMouseEvents, 
		(void*)refcon, 
		NULL );
	if( noErr != err ) {
		printf( "Error installing the mouse event handler: %i.\n", (int)err );
	}
	return err;
}








OSStatus SetupTimer( EventLoopTimerUPP timerHandler, long refcon )
{
	// Define our timer interval (30/second).
	OSStatus err = noErr;
	const EventTime timerInterval = kEventDurationSecond / 30;
	
	// If the handler is NULL, just return.
	if( NULL == timerHandler ) {
		return err;
	}
	
	// Install the handler.
	EventLoopTimerRef ignore;
	err = InstallEventLoopTimer(
		GetMainEventLoop(), 
		kEventDurationNoWait, 
		timerInterval, 
		timerHandler,
		(void*)refcon, 
		&ignore );
	if( noErr != err ) {
		printf( "Error installing timer: %i.\n", (int)err );
		return err;
	}
	
	return noErr;
}


OSStatus SetupAllEventHandlers( 
	EventHandlerUPP appHandler, 
	EventHandlerUPP keyHandler, 
	EventHandlerUPP mouseHandler, 
	EventLoopTimerUPP timerHandler, 
	WindowRef winRef, 
	IBNibRef sNibRef, 
	long refcon
	 )
{
	OSStatus err = noErr;
	// Set up and install the application's main event handler.
	err = SetupAppEventHandler( appHandler, sNibRef, refcon );
	if( noErr != err ) {
		printf( "Error, could not install the Application event handler: %i.\n", (int)err );
		return err;
	}
	
	// Set up and install the keyboard event handler.
	err = SetupKeyboardEventHandler( winRef, keyHandler, refcon );
	if( noErr != err ) {
		printf( "Error, could not install the keyboard event handler: %i.\n", (int)err );
		return err;
	}
	

	// Set up and install the event handler for the mouse.
	err = SetupMouseEventHandler( winRef, mouseHandler, refcon );
	if( noErr != err ) {
		printf( "Error, could not install the mouse event handler: %i.\n", (int)err );
	}
	
	// Set up and install the timer and its corresponding event handler.
	err = SetupTimer( timerHandler, refcon );
	if( noErr != err ) {
		printf( "Error trying to install timer handler: %i.\n", (int)err );
		return err;
	}
	
	return err;
}
