/*
 *  OpenGL_utils.cpp
 *  video_echo3
 *
 *  Created by David Cairns on 3/9/06.
 *
 */

#include "OpenGL_utils.h"


GLenum SetupOpenGLContext( WindowRef winRef, AGLContext *context )
{
	// Specify the attributes for our context's pixel format.
	GLint attribs[] = {
		AGL_RGBA,
		AGL_DOUBLEBUFFER,
		AGL_ACCELERATED,
		AGL_NO_RECOVERY,
		AGL_BUFFER_SIZE, 32,
		AGL_NONE };
	
	// Get a Handle to the main window.
	GDHandle hGD = GetMainDevice();
	
	// Have AGL choose a pixel format for us.
	GLenum result = AGL_NO_ERROR;
	AGLPixelFormat pixFormat = aglChoosePixelFormat( &hGD, 1, attribs );
	if( NULL == pixFormat ) {
		result = aglGetError();
		printf( "Error creating pixel format: %i.\n", (int)result );
		return result;
	}
	
	// Create a context for the pixel format we got from AGL.
	*context = aglCreateContext( pixFormat, NULL );
	result = aglGetError();
	if( AGL_NO_ERROR != result ) {
		printf( "Error creating an OpenGL context: %i.\n", (int)result );
		return result;
	}
	
	// Destroy the pixel format we just created.
	aglDestroyPixelFormat( pixFormat );
	
	// Set the drawable for the context to the window's viewable port.
	GLboolean err;
	err = aglSetDrawable( 
		*context, 
		GetWindowPort( winRef ) );
	if( GL_FALSE == err ) {
		result = aglGetError();
		printf( "Error setting the OpenGL context's drawable: %i.\n", (int)result );
		return result;
	}
	
	// Set the context we've just created as the current one.
	err = aglSetCurrentContext( *context );
	if( GL_FALSE == err ) {
		result = aglGetError();
		printf( "Error setting new OpenGL context to current: %i.\n", (int)result );
		return result;
	}
	
	// Set the swap interval to synchronize with the screen to avoid flicker.
	const GLint swap_interval = 1;
	aglSetInteger( *context, AGL_SWAP_INTERVAL, &swap_interval );
	
	return AGL_NO_ERROR;
}


OSStatus MakeOpenGLWindow( 
	WindowRef *winRef, 
	Rect winBounds, 
	AGLContext *context )
{
	// First, let's set up a new window.
	//	Specify the attributes for the window.
	WindowAttributes windowAttrs = 
		kWindowCloseBoxAttribute
		| kWindowCollapseBoxAttribute
		| kWindowStandardHandlerAttribute
		| kWindowInWindowMenuAttribute;
	
	//  Create the window.
	OSStatus err;
	err = CreateNewWindow( 
		kDocumentWindowClass, 
		windowAttrs, 
		&winBounds, 
		winRef );
	if( noErr != err ) {
		printf( "Error creating new window: %i.\n", (int)err );
		return err;
	}
	
	//  Display the window on the screen.
	ShowWindow( *winRef );
	
	// Now set up an OpenGL context in the window.
	GLenum glErr;
	glErr = SetupOpenGLContext( *winRef, context );
	if( AGL_NO_ERROR != glErr ) {
		printf( "Error creating OpenGL context.\n" );
		return -1;
	}
	
	return noErr;
}


void SetupOpenGLDefaults( Rect portBounds )
{
	// We just want to draw things in 2D!
    glEnable( GL_ALPHA_TEST );
    glDisable( GL_DEPTH_TEST );
	glDisable( GL_BLEND );
	
    glDisable( GL_SCISSOR_TEST );
    glDisable( GL_DITHER );
    glDisable( GL_CULL_FACE );
	
	// Make sure no colors or other components are being masked-out.
    glColorMask( GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE );
    glStencilMask( 0 );
	
	// Set the rendering hint to "fastest", because this app is really
	// performance-intensive.
    glHint( GL_TRANSFORM_HINT_APPLE, GL_FASTEST );
	
	GLsizei screenWidth = (GLdouble)( portBounds.right - portBounds.left );
	GLsizei screenHeight = (GLdouble)( portBounds.bottom - portBounds.top );
	// Setup the matrices we'll be using.
	// We want a basic orthographic projection, because we are doing
	// strict 2D work.
	glViewport( 
		0, 0, 
		screenWidth, 
		screenHeight );
	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluOrtho2D( -1, 33, -1, 25 );
	
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	
	// Set the "default" per-pixel parameters.
    glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
}


unsigned SetupTextureForImageBuffer( ImageBuffer *texBuffer, GWorldPtr graphicsWorld )
{
	printf("SetupTextureForImageBuffer()\n");
	
	// If the user has passed in a GWorldPtr, then the ImageBuffer
	// we are going to be creating a texture from is using
	// that GWorld, and we must tell the GWorld to lock its
	// base address before accessing any of the pixel data directly
	// from it.
	
	// We want to use textures (preferably non-power-of-two).
	if( 0 == OpenGLExtensionIsSupported( "GL_EXT_texture_rectangle" ) ) {
		printf( "Error: This application requires GL_EXT_texture_rectangle!\n" );
	}
	glEnable( GL_TEXTURE_RECTANGLE_EXT );
	
	// Get the "name" for the texture from the GL server.
	GLuint texture_name;
	glGenTextures( 1, &texture_name );
	
	// Tell OpenGL that we want to use the texture we just created.
	glBindTexture( GL_TEXTURE_RECTANGLE_EXT, texture_name );
	
	// Try to store the texture data locally (in our GWorld), 
	// instead of on the graphics card.
	if( 0 != OpenGLExtensionIsSupported( "GL_APPLE_client_storage" ) ) {
		glPixelStorei( GL_UNPACK_CLIENT_STORAGE_APPLE, GL_TRUE );
		glTexParameterf( GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_PRIORITY, 0.0f );
	}
	else {
		printf( "Note -- We can't store textures locally.  This will be SLOW!\n" );
	}
	
	// Finish setting up textures.
	glTexParameteri( GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	
	// If the user supplied a GWorldPtr, we must lock its base address first.
	if( NULL != graphicsWorld ) {
		LockImageBuffer( texBuffer, graphicsWorld );
	}
	// Submit the texture for our ImageBuffer structure.
	int width = texBuffer->bounds.right - texBuffer->bounds.left + 4;
	int height = texBuffer->bounds.bottom - texBuffer->bounds.top;
	glTexImage2D( 
		GL_TEXTURE_RECTANGLE_EXT, 
		0, 
		GL_RGBA8, 
		width, height, 
		0, 
		GL_BGRA_EXT, 
		( OSBigEndian == OSHostByteOrder() ? GL_UNSIGNED_INT_8_8_8_8_REV : GL_UNSIGNED_INT_8_8_8_8 ), 
		texBuffer->buffer );
	// If the user supplied a GWorldPtr, we must unlock it.
	if( NULL != graphicsWorld ) {
		UnlockImageBuffer( texBuffer, graphicsWorld );
	}
	
	return (unsigned)texture_name;
}


void UpdateGLTexture( ImageBuffer *texBuffer, GWorldPtr graphicsWorld )
{
	// Lock the ImageBuffer structure to the GWorld, so we can safely 
	// access the GWorld's base address.
	if( NULL != graphicsWorld ) {
		LockImageBuffer( texBuffer, graphicsWorld );
	}
	
	// Submit the changed texture to OpenGL.
	// **NOTE: We are assuming use of GL_TEXTURE_RECTANGLE_EXT!!!  --DRC
	int width = texBuffer->bounds.right - texBuffer->bounds.left + 4;
	int height = texBuffer->bounds.bottom - texBuffer->bounds.top;
	glTexSubImage2D( 
		GL_TEXTURE_RECTANGLE_EXT, 
		0, 
		0, 0, 
		width, height, 
		GL_BGRA_EXT, 
		( OSBigEndian == OSHostByteOrder() ? GL_UNSIGNED_INT_8_8_8_8_REV : GL_UNSIGNED_INT_8_8_8_8 ), 
		texBuffer->buffer );
	
	// Unlock the GWorld, because we are done with it.
	if( NULL != graphicsWorld ) {
		UnlockImageBuffer( texBuffer, graphicsWorld );
	}
}


/*
 This is SGI sample code taken directly from OpenGL.org:
 http://www.opengl.org/developers/code/features/OGLextensions/OGLextensions.html
 ( Thanks to QTValuePak for this function! )
 */
int OpenGLExtensionIsSupported( const char* extension )
{
	const GLubyte *extensions = NULL;
	const GLubyte *start;
	GLubyte *where, *terminator;

	// Extension names should not have spaces.
	where = (GLubyte *)strchr( extension, ' ' );
	if( 0 != where || '\0' == *extension ) {
		return 0;
	}
	extensions = glGetString( GL_EXTENSIONS );
	// It takes a bit of care to be fool-proof about parsing the
	// OpenGL extensions string. Don't be fooled by sub-strings, etc.
	start = extensions;
	for( ; ; ) {
	where = (GLubyte *)strstr( (const char *)start, extension );
	if( 0 == where ) {
		break;
	}
	terminator = where + strlen( extension );
	if( where == start || ' ' == *(where - 1) )
		if( ' ' == *terminator || '\0' == *terminator ) {
			return 1;
		}
		start = terminator;
	}
	return 0;
}
