/*
 *  QT_utils.c
 *  video_echo3
 *
 *  Created by David Cairns on 3/9/06.
 *
 */

#include "QT_utils.h"


OSErr SetupImageDecompressionSequence( 
	ImageDecompressionSequence	*seq, 
	ImageBuffer					*imageBuffer, 
	WindowRef					windowRef, 
	GWorldPtr					graphicsWorld, 
	SGChannel					channel,
	Ptr							dataPtr, 
	long						dataLength )
{
	OSErr err;
	// Get a Handle to an ImageDescription structure.  This will provide us with some
	// information about the frame that has been passed into this function.
	ImageDescriptionHandle imageDesc = (ImageDescriptionHandle)NewHandle( 0 );
	err = SGGetChannelSampleDescription( channel, (Handle)imageDesc );
	if( noErr != err ) {
		printf( "Error getting an ImageDescription for a frame of video: %i.\n", err );
		return err;
	}
	
	// Get the bounds of the incoming video frame.
	Rect sourceRect = { 0, 0 };
	sourceRect.right = (*imageDesc)->width;
	sourceRect.bottom = (*imageDesc)->height;
	
	// Create a scaling matrix, to scale the video frame from its original dimensions
	// to those stored in our ImageBuffer data structure.
	MatrixRecord scaleMatrix;
	RectMatrix( &scaleMatrix, &sourceRect, &(imageBuffer->bounds) );
	
	// Begin the process of decompressing a sequence of frames into our GWorld.
	err = DecompressSequenceBeginS( 
		&(seq->decompressSeq),
		imageDesc,
		dataPtr,
		dataLength,
		graphicsWorld,
		NULL,						// Specifies that we are drawing offscreen.
		NULL,						// Specifies that we want the whole source image.
		&scaleMatrix,
		srcCopy,
		(RgnHandle)NULL,			// Don't mask the data.
		0,
		codecNormalQuality,
		bestSpeedCodec );
	if( noErr != err ) {
		printf( "Error setting up decompression sequence: %i.\n", (int)err );
		DisposeHandle( (Handle)imageDesc );
		return err;
	}
	
	// Cleanup our ImageDescriptionHandle.
	DisposeHandle( (Handle)imageDesc );
	
	// Specify that this sequence has been initialized.
	seq->initialized = TRUE;
	return noErr;
}


void DestroyImageSequence( ImageDecompressionSequence *imageSequence )
{
	// This function is just a wrapper for the Carbon call.
	// It exists, in case we want to do any extra work when we destroy
	// an ImageDecompressionSequence structure.
	CDSequenceEnd( imageSequence->decompressSeq );
}



void InitImageBuffer( 
	Rect		bounds, 
	ImageBuffer	*imageBuffer, 
	unsigned	numComponents, 
	unsigned	elementSize )
{
	printf ("InitImageBuffer0()\n");
	// Set the fields of the ImageBuffer based on the bounds we've supplied.
	imageBuffer->bytesPerRow = elementSize * ( bounds.right - bounds.left );
	imageBuffer->bounds = bounds;
	
	// Specify how many elements there are, and how large each element is.
	imageBuffer->elemSize = elementSize;
	imageBuffer->numElems = numComponents 
		* ( bounds.right - bounds.left + 4)
		* ( bounds.bottom - bounds.top );
	// Allocate the ImageBuffer's actual buffer.
	imageBuffer->buffer = malloc( imageBuffer->elemSize * imageBuffer->numElems );
		
	printf ("%d %d\n", imageBuffer->buffer, imageBuffer->elemSize * imageBuffer->numElems );
	
	// Clear the image buffer's contents.
	unsigned bufferIndex;
	unsigned char *theBuffer = imageBuffer->buffer;
	for( bufferIndex = 0; bufferIndex < imageBuffer->numElems; bufferIndex++ ) {
		theBuffer[bufferIndex] = 0;
	}
}


OSStatus InitImageBufferWithGWorld( 
	GWorldPtr	*graphicsWorld, 
	Rect		bounds, 
	ImageBuffer *imageBuffer, 
	unsigned	numComponents, 
	unsigned	elementSize )
{
	printf ("InitImageBufferWithGWorld()\n");
	
	// Initialize the GWorld, telling it to use our buffer.
	OSStatus err;
	// Set the fields of the ImageBuffer based on the bounds we've supplied.
	imageBuffer->bytesPerRow = elementSize * ( bounds.right - bounds.left );
	imageBuffer->bounds = bounds;
	// We will assume that the buffer is of either float or int format.
	imageBuffer->elemSize = elementSize;
	imageBuffer->numElems = numComponents 
		* ( bounds.right - bounds.left ) 
		* ( bounds.bottom - bounds.top );
	
	// We will set this pointer value when we call the LockImageBuffer function.
	imageBuffer->buffer = NULL;
	
	
	// Create a new offscreen GWorld that our Sequence Grabber may draw
	// its frames into.
	err = QTNewGWorld(
					  graphicsWorld, 
					  k32ARGBPixelFormat, 
					  &bounds, 
					  0, 
					  NULL, 
					  0 );

	
	if( noErr != err ) {
		printf( "Error trying to create a new GWorld for the sequence grabber: %i.\n", (int)err );
		return err;
	}

	
	
	return noErr;
}


void DestroyImageBuffer( ImageBuffer *imageBuffer )
{
	printf("DestroyImageBuffer() %ld\n", imageBuffer);
	
	if (imageBuffer)
	{
		// Deallocate the contents of the ImageBuffer passed in.
		free( imageBuffer->buffer );
	}
}


void LockImageBuffer( ImageBuffer *theImageBuffer, GWorldPtr graphicsWorld )
{
	if( NULL == graphicsWorld ) {
		printf( "Note: Trying to lock NULL as GWorldPtr.\n" );
	}
	// Lock the GWorld's pixels.
	PixMapHandle pPixMap = GetGWorldPixMap( graphicsWorld );
	if( 0 == LockPixels( pPixMap ) ) {
		printf( "Error trying to lock the base address of the pixmap.\n" );
		return;
	}
	
	// Tell our buffer to use the GWorld's pixels.
	theImageBuffer->buffer = (unsigned char*)GetPixBaseAddr( pPixMap );
}


void UnlockImageBuffer( ImageBuffer *theImageBuffer, GWorldPtr graphicsWorld )
{
	if( NULL == graphicsWorld ) {
		printf( "Note: Trying to lock NULL as GWorldPtr.\n" );
	}
	// Unlock the GWorld's PixMap.
	PixMapHandle pPixMap = GetGWorldPixMap( graphicsWorld );
	UnlockPixels( pPixMap );
	// Invalidate the image buffer that has been locked to this GWorld,
	// so we don't (and can't) accidentally try to access it while it
	// is not locked.
	theImageBuffer->buffer = NULL;
}


OSStatus SetupSequenceGrabber( 
	SeqGrabComponent	*seqGrabber, 
	SGChannel			*sgVideoChannel, 
//	WindowRef			windowRef, 
	GWorldPtr			*graphicsWorld, 
	Rect				imageBounds, 
	SGDataUPP			dataProcCallback, 
	long				refcon )
{
	OSStatus err = couldntGetRequiredComponent;
	
	// Open the default sequence grabber component.
	*seqGrabber = OpenDefaultComponent( SeqGrabComponentType, 0 );  // 0 is subType
	if( NULL == *seqGrabber ) {
		printf( "Error opening default sequence grabber.\n" );
		return err;
	}
	
	
	
	// Initialize the sequence grabber component.
	err = SGInitialize( *seqGrabber );
	if( noErr != err ) {
		printf( "Error initializing sequence grabber: %i.\n", (int)err );
		CloseComponent( *seqGrabber );
		*seqGrabber = NULL;
		return err;
	}
	
	// Tell the sequence grabber to draw its frames into the GWorld.
	err = SGSetGWorld( *seqGrabber, *graphicsWorld,  GetMainDevice() );
	if( noErr != err ) {
		printf( "Error setting sequence grabber's destination GWorld: %i.\n", (int)err );
		CloseComponent( *seqGrabber );
		return err;
	}
	
	// Tell the SG where to "record" its data.
	// Note that using this function implies performing a "record" operation;
	// however we don't want the SG to actually save any daya. We do this because
	// a preview operation will not call our data processing callback function.
	err = SGSetDataRef( *seqGrabber, 0, 0, seqGrabDontMakeMovie );
	if( noErr != err ) {
		printf( "Error trying to set sequence grabber's data reference: %i.\n", (int)err );
		CloseComponent( *seqGrabber );
		*seqGrabber = NULL;
		return err;
	}
	
	// Set up the channel(s) that our sequence grabber will be using.
	err = MakeSequenceGrabChannel( *seqGrabber, sgVideoChannel, &imageBounds );
	if( noErr != err ) {
		printf( "Error setting up sequence grabber channel.\n" );
		CloseComponent( *seqGrabber );
		return err;
	}
	
	// Tell the SG what to call when it has a frame of video ready.
	err = SGSetDataProc( *seqGrabber, dataProcCallback, refcon );
	if( noErr != err ) {
		printf( "Error setting the sequence grabber's data processing function: %i.\n", (int)err );
		DestroySequenceGrabber( *seqGrabber, *sgVideoChannel );
		return err;
	}
	
	// Tell the SG to prepare itself.
	err = SGPrepare( *seqGrabber, false, true );
	if( noErr != err ) {
		printf( "Error: sequence grabber unable to prepare for record operation: %i.\n", (int)err );
		DestroySequenceGrabber( *seqGrabber, *sgVideoChannel );
		return err;
	}
	
	// Tell the SG to begin the record operation.
	// Remember that we are not actually recording anything to memory, but
	// we have to specify a record operation in order for the data 
	// processing function to be called.
	err = SGStartRecord( *seqGrabber );
	if( noErr != err ) {
		printf( "Error trying to begin the record operation: %i.\n", (int)err );
		DestroySequenceGrabber( *seqGrabber, *sgVideoChannel );
		return err;
	}
	
	return noErr;
}


OSStatus MakeSequenceGrabChannel( 
	SeqGrabComponent seqGrabber, 
	SGChannel *sgVideoChannel, 
	Rect *videoBounds )
{
	printf("MakeSequenceGrabChannel\n");
	OSStatus err;
	// Create a new channel for use with the sequence grabber.
	err = SGNewChannel( seqGrabber, VideoMediaType, sgVideoChannel );
	if( noErr != err ) {
		printf( "Error creating new video channel: %i.\n", (int)err );
		return err;
	}
	
	// Tell the video channel where to draw itself.
	err = SGSetChannelBounds( *sgVideoChannel, videoBounds );
	if( noErr != err ) {
		printf( "Error setting video channel's bounds: %i.\n", (int)err );
		return err;
	}
	
	// Set the channel's usage.
	err = SGSetChannelUsage( *sgVideoChannel, seqGrabRecord );
	if( noErr != err ) {
		printf( "Error attempting to set video channel usage: %i.\n", (int)err );
		return err;
	}
	
	// Tell the sequence grabber that we want to draw offscreen.
	err = SGSetUseScreenBuffer( *sgVideoChannel, FALSE );
	if( noErr != err ) {
		printf( "Error setting sequence grabber to draw itself offscreen: %i.\n", (int)err );
		SGDisposeChannel( seqGrabber, *sgVideoChannel );
		return err;
	}
	
	return noErr;
}


void DestroySequenceGrabber( SeqGrabComponent seq, SGChannel sgChannel )
{
	// Clean up our SGChannel and our sequence grabber.
	if( NULL != seq ) {
		SGDisposeChannel( seq, sgChannel );
		CloseComponent( seq );
		seq = NULL;
	}
}
